; #########################################################################
;
;                             qwtoa
;
; Procedure for converting a 64-bit signed integer from memory to ASCII
;
; Limitation: The st5, st6 and st7 registers on the FPU will be trashed
;
; Return: EAX!=0 and null-terminated ASCII string in memory
;         EAX =0 if FPU unusable and null string in memory
;
; Usage:  invoke qwtoa,lpQW,lpBuf
;
;   where:  
;   lpQW  = pointer to the location of the qword in memory
;   lpBuf = pointer to a memory buffer for the null-terminated ASCII
;           string (a 21-byte buffer is sufficient)
;
; All CPU registers except EAX are preserved
;
;                       by Raymond Filiatreault
;                              May 2005
;                            MASM  syntax
;
; Algorithm:
; The range of a signed qword is +/-9223372036854775807 which has up to
; 19 digits. By dividing by 10, the quotient would then have a maximum of
; only 18 digits which can then be handled with the FBSTP instruction. The
; single-digit remainder of the division is later converted to ASCII and
; appended to the string converted from the stored packed BCD of the quotient.
;
; #########################################################################

;      .386                   ; minimum processor needed for 32 bit
;      .model flat, stdcall   ; FLAT memory model & STDCALL calling
;      option casemap :none   ; set code to case sensitive

; #########################################################################

qwtoa  PROTO :DWORD, :DWORD

.code

qwtoa proc USES ecx esi edi lpQW:DWORD, lpBuf:DWORD

LOCAL tempBCD     :TBYTE
LOCAL oldCW       :WORD
LOCAL newCW       :WORD

      fclex                   ;clear exception flags
      ffree st(7)
      ffree st(6)
      ffree st(5)             ;3 FPU registers required

      fstcw oldCW
      mov   ax,oldCW
      or    ah,03h            ;modify Control Word for maximum precision
      mov   newCW,ax
      fldcw newCW             ;load modified Control Word

      mov   eax,lpQW          ;retrieve address of qword
      fild  qword ptr[eax]    ;load the qword on the FPU
      fld   st                ;copy qword
      mov   eax,10
      push  eax
      fidiv dword ptr[esp]

; The quotient must be truncated to an integer before storing as a packed BCD

      or    newCW,0C00h       ;modify Control Word for truncating
      fldcw newCW             ;load modified Control Word
      frndint                 ;truncate quotient to an integer

      fld   st                ;copy truncated quotient
      fbstp tempBCD           ;store quotient
      fimul dword ptr[esp]    ;multiply it by 10
      fsub                    ;subtract it from original qword
      fabs
      fistp dword ptr[esp]    ;store last digit on the stack and clean FPU
      fldcw oldCW             ;restore original Control Word

      fstsw ax                ;to check for exception flags
      fwait
      shr   al,1              ;check for FPU Invalid operation
      jnc   @F                ;continue if none

      pop   eax               ;clean stack
      xor   eax,eax
      mov   edi,lpBuf
      mov   [edi],al
      ret                     ;return with error code and null string

; Convert packed BCD to ASCII

   @@:
      lea   esi,tempBCD+9     ;point to most significant byte
      mov   edi,lpBuf
      mov   ecx,9
      mov   al,[esi]
      shl   al,1              ;move sign bit to CF
      jnc   @F
      mov   al,"-"
      stosb

; Find first non-zero digit

   @@:
      dec   esi
      cmp   byte ptr[esi],0
      jnz   @F
      dec   ecx
      jnz   @B
      jmp   lastdigit
   @@:
      mov   al,[esi]
      test  al,0F0h           ;check high nibble of 1st non-zero byte
      jz    lownibble

   @@:
      mov   al,[esi]
      push  eax
      shr   al,4              ;shift high nibble to low nibble
      add   al,30h            ;convert it to ASCII
      stosb                   ;store it in string
      pop   eax
      and   al,0Fh            ;keep only low nibble
lownibble:
      add   al,30h
      stosb
      dec   esi
      dec   ecx
      jnz   @B

lastdigit:
      pop   eax               ;retrieve least significant digit of qword
      add   al,30h
      stosw                   ;store it with a terminating 0
      ret

qwtoa endp

;end
